--[[

# TagProcessor.lua

This module parses and evaluates inputs containing tags that are surrounded by 
curly braces. Inputs that are formatted in the manner that is expected by this 
module can be used in addon settings, allowing users greater control over data
formatting and processing.


## Summary

Tags are formatted as a data source, followed by a colon (:), followed by a 
period (.) delimited path from the root of the data source to a single data 
point. Tags can be nested, which can be useful in cases where the resolved 
value of one tag needs to be used to construct part of a path in another tag. 
There can also be more than tag in a given input. Most tags will be formatted 
either `{DataSource:Value}` or `{DataSource:Table.Value}`. 

There are 3 data sources that are currently supported:

- TableField
- Setting
- DataMapping

### TableField
If the input string contains any patterns that reference `TableField` as a 
data source, then the ReplaceTags function should be called as part of a 
ProcessDataContexts routine. Otherwise, the addon may crash.

### Setting
This module does not interact directly with the `config.xml`. If the input 
string contains `Setting` as a data source, then settings should be loaded 
ahead of time. Some settings need to be pre-processed before the module can 
operate on the setting. For example, some settings can be formatted as colon 
separated lists of `key=value` pairs, which must be resolved to a table before 
this module can reference individual keys. 

### DataMapping
If the input string contains `DataMapping` as a data source, then a 
DataMapping.lua file should be created and populated with information related 
to mapping data to/from relevant database fields. 


## Usage

The snippet below returns the value stored in the `AnAddonSetting` setting. 

```lua
TagProcessor:ReplaceTags('{Setting.AnAddonSetting}')
```

The snippet below first evaluates `Transaction.Site` in reference to the 
current transaction. Then it uses that value as a reference to an item in the 
SettingTable and returns the result. In this example, the parent Addon needs 
to have a `Settings` table with a `SettingTable` sub-table. 

```lua
TagProcessor:ReplaceTags('{Setting:SettingTable.{TableField:Transaction.Site}}')
```

The snippet below returns the string 'Transaction Number: (<tn>)', where 
`<tn>` is equal to the transaction number of the current transaction.

```lua
TagProcessor:ReplaceTags('Transaction Number: ({TableField:Transaction.TransactionNumber})')
```

The snippet below returns the value stored in `DataMapping.X.Y`. In this 
example, the addon likely has a DataMapping.lua file, which is commonly used 
to specify string resources or to allow the addon to switch between Atlas 
products. 

```lua
TagProcessor:ReplaceTags('{DataMapping:X.Y}')
```


## Notes

The addon where this was first used had some settings that were semicolon (;) 
separated lists of `key=value` pairs. These settings were parsed into Lua 
tables. These lists of k,v pairs also had `Default` as one of the keys, which
is referenced in the event that 

Default=22101002531549;RBSC=22101002531549;MUDD=22101002930501;MARQ=22101001828219


## .NET Library Requirements

- log4net


## Addon Module Requirements

- Utilities.lua

]]


luanet.load_assembly("System")
luanet.load_assembly("log4net")

TagProcessor = {}
TagProcessor.__index = TagProcessor

TagProcessor.Log = luanet.import_type("log4net.LogManager").GetLogger("AtlasSystems.Addons.TagProcessor")

function TagProcessor:ReplaceTags (input)
    --[[ 
        
        Recursively parses input for items surrounded in curly braces ("{" and 
        "}") and evaluates tokes that match the regex: 
        
        `{(DataMapping|TableField|Setting):[\w\.]+}`
        
    ]]
    local Log = TagProcessor.Log

    if not input or input == "" then
        return ""
    end

    -- Check for sub-tags and evaluate
    Log:DebugFormat("[ReplaceTags] Processing: {0}", input)
    local compoundTag = (input or ""):match("^{(.+)}$") or input

    local leftBraceIndex = compoundTag:find("{")

    if (leftBraceIndex) then
        local braceLevel = 0

        for i = leftBraceIndex, #compoundTag do
            local character = compoundTag:sub(i,i)
            
            -- Check for sub-nested tags
            if     (character == "{") then braceLevel = braceLevel + 1
            elseif (character == "}") then braceLevel = braceLevel - 1
            end

            if braceLevel == 0 then
                local subTag = compoundTag:sub(leftBraceIndex, i)
                Log:DebugFormat("[ReplaceTags] Found Sub Tag: {0}", subTag)
                compoundTag = compoundTag:gsub(subTag, TagProcessor:ReplaceTags(subTag))
                return TagProcessor:ReplaceTags(compoundTag)
                -- Reason for the above early termination with a reprocessed compoundTag. In the example 
                -- Settings:{TableField:X.Y}.{TableField:X.Z}, the X.Y subtag is processed first, yielding 
                -- Settings:XYValue.{TableField:X.Z}, which has to be reprocessed. 
            end 
        end
    end
    
    local tag = compoundTag

    --Remove the ",EscapeQuotes"
    local escapeQuotesIndex = string.find(tag, ",EscapeQuotes")
    if (escapeQuotesIndex and escapeQuotesIndex > 0) then
        tag = string.sub(tag, 1, escapeQuotesIndex - 1)
        Log:DebugFormat("[ReplaceTags] Tag will be escaped: {0}", tag)
    else
        Log:DebugFormat("[ReplaceTags] Tag will not be escaped: {0}", tag)
    end

    local trimmedTag = tag:gsub("^[^:]+:", "")

    if tag:lower():find("^tablefield:") then
        returnValue = TagProcessor:MapFieldValue(trimmedTag)

    elseif tag:lower():find("^datamapping:") then
        returnValue = TagProcessor:MapDataMapping(trimmedTag)

    elseif tag:lower():find("^setting:") then
        returnValue = TagProcessor:MapSettingValue(trimmedTag)

    else
        returnValue = tag

    end

    if (escapeQuotesIndex) then
        returnValue = returnValue:gsub('"', '\"')
    end

    Log:DebugFormat("[ReplaceTags] Tag replacement result: '{0}' => '{1}'", tag, returnValue)
    return returnValue
end


function TagProcessor:MapSettingValue (settingDefinition)
    --[[

        Assumes that input is of the form X.Y or X.
        
        If X.Y : then Settings[X][Y] is returned. If nil, then 
        Settings[X].Default is returned, if nil, then Settings[X] is returned. 
        
        If X : then Settings[X] is returned. 

        If either are nil, then "" is returned.

    ]]
    local Log = TagProcessor.Log

    if not Settings then
        Log:Warn("[MapSettingValue] No Settings found.")
        return ""

    elseif not settingDefinition then
        Log:Warn("[MapSettingValue] No value was provided to MapSettingValue")
        return ""

    end

    Log:DebugFormat("[MapSettingValue] Mapping Settings value: '{0}'", settingDefinition)

    local separatorIndex = string.find(settingDefinition, "%.")
    if (separatorIndex and separatorIndex > 0) then
        
        --[[
            This section should catch all cases where the mapping
            indicates that the setting is formatted as a sub-table.
        ]]

        local table = settingDefinition:sub(1, separatorIndex - 1)
        local field = settingDefinition:sub(separatorIndex + 1)

        if not Settings[table] then
            Log:WarnFormat("[MapSettingValue] There is no setting named '{0}'", table)
            return ""
        end

        local subTable = Settings[table]

        if type(subTable) == "string" then
            subTable = TagProcessor:ParseMappingSetting(subTable)            
            if type(subTable) == "string" then
                Log:WarnFormat("[MapSettingValue] Setting '{0}' is being used as a table, but is not formatted as a semicolon separated list of key=value pairs.", table)
                return subTable
            end
        end

        if subTable[field] then
            return subTable[field]

        elseif subTable.Default then
            return subTable.Default

        else
            Log:WarnFormat("[MapSettingValue] Setting '{0}' has no value for '{1}', and does not have a Default value.", table, field)
            return ""
        end
    else

        --[[
            This section should catch all cases where the mapping
            indicates that the setting is just a string.
        ]]
    
        local setting = Settings[settingDefinition]

        if (not setting) then
            Log:WarnFormat("[MapSettingValue] There is no setting with the name '{0}'", settingDefinition)
            return ""

        elseif type(setting) == 'table' then
            return setting.Default

        else 
            return setting
            
        end
    end
end


function TagProcessor:ParseMappingSetting (setting)
    --[[

        Parses a mapping setting, a semicolon separated list of key/value 
        pairs (k1=v1[;k2=v2[...] ]), and returns the result as k_n, v_n pairs 
        in a table. If there is only a single value, with no = character, then 
        just the value is returned as a string. If there are multiple values 
        with no = character, then the values are returned as an array (a table 
        with only numeric indices). 

    ]]

    local result = {}

    for mapping in string.gmatch(setting, "[^;]+") do
        local indexOfEq = mapping:find("=")

        if indexOfEq then
            result[ mapping:sub(1, indexOfEq - 1) ] = mapping:sub( indexOfEq + 1 )

        else
            if setting:find(";") then
                table.insert(result, mapping)
            else
                result = mapping
            end
        end
    end

    return result
end


function TagProcessor:MapFieldValue (fieldDefinition)
    --[[

        Assumes that input is of the form X.Y. If X is "Setting", then this 
        will try to return the value of GetSetting(Y). Otherwise, this will 
        try to return the value of GetFieldValue(X, Y). 

    ]]
    local Log = TagProcessor.Log 
    

    if fieldDefinition then
        Log:DebugFormat("[MapFieldValue] Mapping FieldValue for {0}", fieldDefinition)
        local separatorIndex = string.find(fieldDefinition, "%.")

        if (separatorIndex and separatorIndex > 0) then
            local table = string.sub(fieldDefinition, 1, separatorIndex - 1)
            local field = string.sub(fieldDefinition, separatorIndex + 1)

            local value = nil
            value = GetFieldValue(table, field)

            if not value then
                Log:Debug("[MapFieldValue] Replacement value is null.")
                value = ""
            elseif Utilities:IsType(value, "System.DateTime") then
                value = value:ToString("yyyy-MM-dd")
                Log:DebugFormat("[MapFieldValue] Value Date: {0}", value)
            end

            return value or ""
        else
            return fieldDefinition
        end
    end

    return ""
end


function TagProcessor:MapDataMapping (fieldDefinition)
    --[[

        Tries to return the data mapping for fieldDefinition. Requires a 
        DataMapping table. 

    ]]

    local Log = TagProcessor.Log 
    

    if (fieldDefinition) then

        Log:DebugFormat("[MapDataMapping] Mapping DataMapping for {0}", fieldDefinition)

        local dataMappingIndex = fieldDefinition:find(":")

        if (dataMappingIndex and dataMappingIndex > 0) then

            dataMappingObject = fieldDefinition:sub(1, dataMappingIndex - 1)
            Log:DebugFormat("[MapDataMapping] Object: {0}", dataMappingObject)

            local dataMappingObjectIndex = dataMappingObject:find(".", 1, true)
            local dataMappingArray = string.sub(dataMappingObject, dataMappingObjectIndex + 1)
            Log:DebugFormat("[MapDataMapping] Array: {0}", dataMappingArray)

            local dataMappingLookupTag = string.sub(fieldDefinition, dataMappingIndex + 1)
            Log:DebugFormat("[MapDataMapping] Lookup Tag: {0}", dataMappingLookupTag)

            dataMappingLookup = TagProcessor:ReplaceTags(dataMappingLookupTag)
            Log:DebugFormat("[MapDataMapping] Getting DataMapping from {0} for {1}", dataMappingObject, dataMappingLookup)

            local value = DataMapping[dataMappingArray][dataMappingLookup]

            if value then
                Log:DebugFormat("[MapDataMapping] Found datamapping value on {0} for {1}: {2}", dataMappingObject, dataMappingLookup, value)
            else
                Log:DebugFormat("[MapDataMapping] Datamapping value not found on {0} for {1}. Using Default.", dataMappingObject, dataMappingLookup)
                value = DataMapping.Defaults[dataMappingArray]
            end

            if value then
                return value
            else
                Log:DebugFormat("Using blank value for {0}", fieldDefinition)
                return ""
            end

        else
            Log:Debug("Could not determine the lookup value for the datamapping")
        end
    else
        Log:Debug("DataMapping not provided")
    end
    return ""
end
